require_relative 'registration_builder'

module CrystalClear
  class Resolver

    # @return [<? extends class_name>]
    def get(class_name)
      result = storage[class_name]
      raise ClassNotRegisteredError.new("No information about class: #{class_name}") unless result
      result.instantiate
    end

    # @param [CrystalClear::RegistrationBuilder] builder
    # @return [CrystalClear::Resolver]
    def register(builder)
      raise ArgumentError, "#{builder.class} is not a CrystalClear::RegistrationBuilder" unless builder.is_a?(CrystalClear::RegistrationBuilder)
      builder.build(add_procedure)
      self
    end

    # @param [Class] class_name
    # @return [CrystalClear::Resolver]
    def register_class(class_name)
      raise ArgumentError, "#{class_name.class} is not a Class" unless class_name.is_a?(Class)
      register(CrystalClear::RegistrationBuilder.for(class_name))
    end

    # @param [Class] class_name
    # @return [CrystalClear::Resolver]
    def register_as_singleton(class_name)
      raise ArgumentError, "#{class_name.class} is not a Class" unless class_name.is_a?(Class)
      register(CrystalClear::RegistrationBuilder.for(class_name).resolves_to_singleton)
    end

    # @param [Class] provider_class
    # @param [Class] class_name
    # @return [CrystalClear::Resolver]
    def register_provider(provider_class, class_name)
      raise ArgumentError, "#{provider_class} is not a Class" unless provider_class.is_a?(Class)
      raise ArgumentError, "#{provider_class} is not a Provider" unless provider_class < CrystalClear::Provider
      raise ArgumentError, "#{class_name} is not a Class" unless class_name.is_a?(Class)
      register(CrystalClear::RegistrationBuilder.for(class_name).resolves_to_provider(provider_class))
    end

    # @param [Class] provider_class
    # @param [Class] class_name
    # @return [CrystalClear::Resolver]
    def register_singleton_provider(provider_class, class_name)
      raise ArgumentError, "#{provider_class} is not a Class" unless provider_class.is_a?(Class)
      raise ArgumentError, "#{provider_class} is not a Provider" unless provider_class < CrystalClear::Provider
      raise ArgumentError, "#{class_name} is not a Class" unless class_name.is_a?(Class)
      register(CrystalClear::RegistrationBuilder.for(class_name).resolves_to_singleton_provider(provider_class))
    end

    private

    def storage
      @storage ||= {}
    end

    def add_procedure
      @add_procedure ||= proc do |classes, instantiator, class_name|
        classes.each do |c|
          if storage[c].nil?
            storage[c] = instantiator
          else
            raise ClassAlreadyRegisteredError.new("Class #{c} already registered for class: #{class_name}")
          end
        end
      end
    end

  end

  class ClassAlreadyRegisteredError < RuntimeError
  end

  class ClassNotRegisteredError < RuntimeError
  end

end
